try:
    import sqlite3 # Python 2.5 and up
except ImportError:
    from pysqlite2 import dbapi2 as sqlite3

_database_defs = """\
create table config(name primary key, value);
create table dailytally(day date, path, count int, primary key (day, path));

create table weektally(path primary key, type, count int);
create index weekcountidx on weektally (count);
create index weektypeidx on weektally (type);

create table monthtally(path primary key, type, count int);
create index monthcountidx on monthtally (count);
create index monthtypeidx on monthtally (type);

create table yeartally(path primary key, type, count int);
create index yearcountidx on yeartally (count);
create index yeartypeidx on yeartally (type);
"""

class ConnectionWrapper(object):
    """Wrapper for SQLite connection

    The wrapper adds some convenience methods, and makes sure all tables
    are created.
    
    >>> conn = ConnectionWrapper(':memory:')
    >>> conn.execute('select * from dailytally;').fetchall()
    []
    
    The setConfig and getConfig methods can be used to store some simple
    name/value pairs.
    
    >>> conn.getConfig('foo', u'default')
    u'default'
    >>> conn.setConfig('foo', 'bar')
    >>> conn.getConfig('foo', 'default')
    u'bar'
    
    Returned values can be coerced to any type recognized by sqlite3
    
    >>> import datetime
    >>> conn.setConfig('today', datetime.date.today())
    >>> res = conn.getConfig('today', type='date')
    >>> res.day == datetime.date.today().day
    True
    
    """
    def __init__(self, database):
        detect_types = sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES
        self._conn = sqlite3.connect(database, detect_types=detect_types)
        # "pragma user_version" causes a segfault, more's the pity
        if not self._conn.execute('pragma table_info(config)').fetchone():
            self._create_schema()
            
    def __getattr__(self, attr):
        return getattr(self._conn, attr)
    
    def _create_schema(self):
        self._conn.executescript(_database_defs)
        self._conn.commit()
    
    def getConfig(self, prop, default=None, type=None):
        """Get a configuration property, or default if not defined

        Type can be used to force coercion.
        
        """
        statement = 'select value from config where name = ?'
        if type is not None:
            statement = ('select value as "v [%s]" from config '
                         'where name = ?' % type)
        res = self._conn.execute(statement, (prop,)).fetchone()
        if not res:
            return default
        return res[0]
    
    def setConfig(self, prop, value):
        """Set a configuration property to value"""
        _marker = []
        if self.getConfig(prop, _marker) is _marker:
            stmt = 'insert into config (name, value) values (:prop, :value)'
        else:
            stmt = 'update config set value=:value where name=:prop'
        self._conn.execute(stmt, locals())
        self._conn.commit()
        
_conn = {}
def getConnection(config):
    """Returns the database connection for the given configuration"""
    global _conn
    conn = _conn.get(config.database)
    if conn is None:
        conn = _conn[config.database] = ConnectionWrapper(config.database)
    return conn

def _test(module=None):
    import doctest
    return doctest.testmod(m=module)

if __name__ == '__main__':
    _test()
