import datetime
import md5
import time
import re

class LogParser(object):
    """Parse a HTTP log, yielding path and date information
    
    For now, only supports Apache Combined format
    
    Given a file object, this class will act as a generator, yielding 
    (timestamp, path) tuples.
    
      >>> import StringIO, datetime
      >>> log = '\\n'.join((
      ... '192.168.120.75 - - [31/Jan/2006:10:39:40 +0100] "GET / HTTP/1.1" '
      ... '200 1096 "-" "Mozilla/4.0 (compatible; MSIE 6.0)"',
      ... '192.168.120.75 - - [31/Jan/2006:10:39:40 +0100] "GET '
      ... '/icons/blank.gif HTTP/1.1" 200 148 "http://boomslang/" '
      ... '"Mozilla/4.0 (compatible; MSIE 6.0)"',))
      >>> parser = LogParser(StringIO.StringIO(log))
      >>> results = list(parser)
      >>> len(results)
      2
      >>> results[0]
      (datetime.datetime(2006, 1, 31, 9, 39, 40), '/')
      >>> results[1]
      (datetime.datetime(2006, 1, 31, 9, 39, 40), '/icons/blank.gif')
      
    The parser keeps track of the last line it parsed, so that it won't report
    the same lines again since last time:
    
      >>> position = parser.position
      >>> for key in ('offset', 'checksum', 'timestamp'):
      ...     position[key]
      115
      '7bc170825f818f7db391435e7aa755b5'
      datetime.datetime(2006, 1, 31, 9, 39, 40)
      
    This position marker can then be used later to skip over old data and
    only report lines that have been added since the previous report:
    
      >>> log += (
      ... '\\n192.168.120.75 - - [31/Jan/2006:10:39:41 +0100] "GET '
      ... '/icons/folder.gif HTTP/1.1" 200 225 "http://boomslang/" '
      ... '"Mozilla/4.0 (compatible; MSIE 6.0)"')
      >>> parser = LogParser(StringIO.StringIO(log), position=position)
      >>> results = list(parser)
      >>> results
      [(datetime.datetime(2006, 1, 31, 9, 39, 41), '/icons/folder.gif')]
      
    If the logfile seek fails, records are skipped based on the position
    timestamp handed in:
    
      >>> log = log.splitlines(True)[0] + log
      >>> parser = LogParser(StringIO.StringIO(log), position=position)
      >>> results = list(parser)
      >>> results
      [(datetime.datetime(2006, 1, 31, 9, 39, 41), '/icons/folder.gif')]
      >>> parser.position['offset']
      375
      
    If the logfile then doesn't change, no new lines are returned:
    
      >>> newparser = LogParser(StringIO.StringIO(log), 
      ...                       position=parser.position)
      >>> len(list(newparser))
      0
      >>> newparser.position == parser.position
      True
      
    """
    _format = re.compile(r'''
        \S+\ \S+\ [^[]+\             # domain name or ip, other, and username
        \[(?P<timestamp>\S+\ \S+)\]\ # timestamp
        "(?P<method>\S+)\            # method (GET, POST, etc)
         (?P<path>\S+)\              # path requested
         [^"]+"\                     # HTTP version of request
        (?P<response>[\d|-]+)\       # HTTP response code
        [\d|-]+\ ".*?"\ "[^"]*"      # size in bytes, referrer, useragent 
    ''', re.X)
        
    _methods = ('GET',)
    _responses = (200, 203, 206, 304)
    _skip_until = None
    
    def __init__(self, log, position=None):
        self._log = log
        self.position = dict(offset=None, checksum=None, timestamp=None)
        if position is not None:
            self.position.update(position)
            self._seek_to_old_position()
        
    def _parseTimestamp(self, timestamp):
        """Convert a log timestamp to a datetime instance.
        
        Returns None when it cannot parse the timestamp

        Expects a 'dd/Mon/yyyy:hh:mm:ss +tttt' formatted string, and returns a
        datetime.datetime instance in UTC:
        
          >>> parser = LogParser(None)
          >>> parser._parseTimestamp('31/Jan/2006:10:39:40 +0100')
          datetime.datetime(2006, 1, 31, 9, 39, 40)
          
        """
        timestamp, tzoffset = timestamp.split(' ')
        try:
            localtime = time.strptime(timestamp, '%d/%b/%Y:%H:%M:%S')
            localtime = time.mktime(localtime)
            tzdelta = (int(tzoffset[1:3]) * 60) + int(tzoffset[3:5])
            if tzoffset[0] == '-':
                tzdelta *= -1
            localtime -= tzdelta * 60
        except (OverflowError, ValueError):
            return None
        timestamp = datetime.datetime.fromtimestamp(localtime)
        return timestamp
    
    def _seek_to_old_position(self):
        offset, checksum = self.position['offset'], self.position['checksum']
        if not (offset and checksum):
            self.position.update(dict(offset=None, checksum=None))
            self._skip_until = self.position['timestamp']
            return
        try:
            self._log.seek(offset)
        except IOError:
            # Possible seek beyond end
            self.position.update(dict(offset=None, checksum=None))
            self._skip_until = self.position['timestamp']
        else:
            line = self._log.readline().strip()
            if md5.new(line).hexdigest() != checksum:
                # Not the same line anymore, start from 0
                self._log.seek(0)
                self.position.update(dict(offset=None, checksum=None))
                self._skip_until = self.position['timestamp']
        
    def __iter__(self):
        nextoffset = self._log.tell()
        while True:
            # Use readline instead of 'for line in log' to avoid the read-ahead
            # buffer used by the iterator; doing so will mess up the .tell()
            # calls so important to this class.
            line = self._log.readline()
            if not line:
                return
            line = line.strip()
            match = self._format.match(line)
            if not match:
                continue
            if match.group('method') not in self._methods:
                continue
            try:
                response = int(match.group('response'))
            except ValueError:
                continue
            else:
                if response not in self._responses:
                    continue
            timestamp = self._parseTimestamp(match.group('timestamp'))
            
            if not timestamp:
                continue
            
            self.position.update(dict(
                offset=nextoffset, timestamp=timestamp,
                checksum=md5.new(line).hexdigest()))
            # We need to remember the position before reading the next line
            # to be able to seek to it later.
            nextoffset = self._log.tell()

            if self._skip_until and timestamp <= self._skip_until:
                continue
            self._skip_until = None
            
            yield timestamp, match.group('path')

def _test(module=None):
    import doctest
    return doctest.testmod(m=module)

if __name__ == '__main__':
    _test()
