import urllib
try:
    # Python 2.5 includes ElementTree
    import xml.etree.cElementTree as ElementTree
except ImportError:
    try:
        # Fall back to cElementTree elsewhere
        import cElementTree as ElementTree
    except ImportError:
        try:
            # Try python-only ElementTree in the Py2.5 location
            from xml.etree import ElementTree
        except ImportError:
            # Python-only ElementTree everywhere else
            from elementtree import ElementTree

class LinksQuery(object):
    """Retrieve links-to-tally info from a URL
    
    Given a url, this class will act as an iterator, yielding (path, type)
    tuples.
    
    To illustrate, we create a data: url containing a small tallylinks XML
    document:
    
      >>> import base64
      >>> links = '''\\
      ... <?xml version="1.0" encoding="utf-8"?>
      ... <!DOCTYPE tallylinks>
      ... <tallylinks>
      ... <link path="/foo" type="spam" />
      ... <link path="/bar" type="eggs" />
      ... </tallylinks>'''
      >>> url = 'data:text/xml;base64,' + base64.encodestring(links).strip()
      
    Using this url, we can then use the LinksQuery class to iterate over the
    information contained within:
    
      >>> lquery = LinksQuery(url)
      >>> for path, type in lquery:
      ...     print path, type
      /foo spam
      /bar eggs
    
    """
    
    def __init__(self, url):
        self._url = url
        
    def __iter__(self):
        source = urllib.urlopen(self._url)
        for event, elem in ElementTree.iterparse(source):
            if elem.tag == 'link':
                yield elem.attrib['path'], elem.attrib['type']
            elem.clear()

def _test(module=None):
    import doctest
    return doctest.testmod(m=module)

if __name__ == '__main__':
    _test()
