import datetime

class LinkTallies(object):
    """Class to manage link tallies.
    
    Counts are kept for moving windows of various lenghts, and are updated
    when instanciated by substracting tallies falling outside of the window.
    
    Some setup is in order to demonstrate this:
    
      >>> import db, datetime
      >>> today = datetime.datetime.utcnow().date()
      >>> day = datetime.timedelta(1)
      >>> week = datetime.timedelta(7)
      >>> month = datetime.timedelta(30)
      >>> year = datetime.timedelta(  365)
      >>> conn = db.ConnectionWrapper(':memory:')
      >>> exc =  conn.execute
      >>> for i in range(366):
      ...     res = exc('insert into dailytally values (?, ?, ?)',
      ...               ((today - year) + ((i - 2) * day), '/foo', 365 - i))
      >>> for i in range(31):
      ...     res = exc('insert into dailytally values (?, ?, ?)',
      ...               ((today - month) + ((i - 2) * day), '/bar', 30 - i))
      >>> res = conn.executescript('''\\
      ... insert into weektally  values('/foo', 'spam', 28);
      ... insert into monthtally values('/foo', 'spam', 465);
      ... insert into yeartally  values('/foo', 'spam', 66795);
      ... insert into weektally  values('/bar', 'spam', 28);
      ... insert into monthtally values('/bar', 'spam', 465);
      ... insert into yeartally  values('/bar', 'spam', 465);''')
      >>> conn.commit()
    
    Now, if the last tally took place 2 days ago, instanciating LinkTallies 
    will update the tallies by substracting the 2 days outside of the past
    week, month and year periods:
    
      >>> conn.setConfig('tallydate', today - (2 * day))
      >>> tallies = LinkTallies(conn)
      >>> exc('select count from weektally order by path').fetchall()
      [(17,), (17,)]
      >>> exc('select count from monthtally order by path').fetchall()
      [(408,), (408,)]
      >>> exc('select count from yeartally order by path').fetchall()
      [(465,), (66068,)]
      >>> exc('select * from dailytally where day <= ?',
      ...     (today - year,)).fetchall()
      []
      >>> exc('select count(*) from dailytally where day = ?',
      ...     (today - year + day,)).fetchone()
      (1,)
    
    Tallies are only kept for registered paths; use updatePaths to ensure
    that paths will correctly be tallied, and old paths are removed.
    
      >>> paths = (('/foo', 'spam'), ('/baz', 'eggs'))
      >>> tallies.updatePaths(paths)
      >>> exc("select * from dailytally where path = '/bar'").fetchall()
      []
      >>> exc('select distinct path from yeartally order by path').fetchall()
      [(u'/baz',), (u'/foo',)]
      >>> exc("select count from yeartally where path = '/baz'").fetchone()
      (0,)
    
    An instance can then be updated with dated tallies:
    
      >>> tallies.tallyPath('/foo', today)
      >>> tallies.tallyPath('/foo', today - week, 5)
      >>> tallies.tallyPath('/baz', today - week, 5)
      >>> tallies.tallyPath('/baz', today, 10)
      >>> exc('select count from weektally order by path').fetchall()
      [(10,), (18,)]
      >>> exc('select count from monthtally order by path').fetchall()
      [(15,), (414,)]
      >>> exc('select count from yeartally order by path').fetchall()
      [(15,), (66074,)]
      >>> res = exc("select day, count from dailytally where path = '/baz' "
      ...           "order by day").fetchall()
      >>> res == [(today - week, 5), (today, 10)]
      True
      >>> exc("select count from dailytally where path = '/foo' "
      ...     "and day = ?", (today - week,)).fetchone()
      (10,)
    
    """
    periods = dict(year=365, month=30, week=7)
    
    def __init__(self, conn):
        self._conn = conn
        self._substractOldTallies()
        
    def _substractOldTallies(self):
        """Substract outdated tallies from the database"""
        today = datetime.datetime.utcnow().date()
        tallydate = self._conn.getConfig('tallydate', type='date')
        if tallydate is None:
            # Never tallied; store initial value of today
            self._conn.setConfig('tallydate', today)
            self._conn.commit()
            return
        if today <= tallydate:
            # last tally date today or in the future (timezone error?)
            return
        earliest = tallydate
        for period, length in self.periods.items():
            to_date = today - datetime.timedelta(length)
            from_date = tallydate - datetime.timedelta(length - 1)
            self._substractTalliesFor(period, from_date, to_date)
            earliest = min(earliest, to_date)
        # Remove old daily tallies
        self._conn.execute('delete from dailytally where day <= ?', 
                           (earliest,))
        self._conn.setConfig('tallydate', today)
        self._conn.commit()
            
    def _substractTalliesFor(self, period, from_date, to_date):
        """Substract all tallies from given period for period data"""
        table = '%stally' % period
        temptable = 'old_counts_%s' % table
        self._conn.execute('''\
            create temp table %s as 
            select path, sum(count) as oldcount from dailytally
            where day between ? and ? group by path;
            ''' % temptable, (from_date, to_date))
        self._conn.execute('''\
            update %(table)s set count = count - ifnull((
              select oldcount from %(temptable)s 
              where %(temptable)s.path = %(table)s.path), 0)
            ''' % locals())
        
    def updatePaths(self, paths):
        """Update tallied paths
        
        'paths' is a iterable yielding (path, type) tuples. Any path in the 
        tallies tables not in this iterable is discarded, and any as yet 
        untallied path is added.
        
        """
        self._conn.execute('create temp table update_paths (path, type)')
        self._conn.executemany('insert into update_paths values (?, ?)', paths)
        self._conn.execute('delete from dailytally where path not in '
                           '(select path from update_paths)')
        for period in self.periods:
            self._conn.executescript('''\
                delete from %(period)stally where path not in (
                    select path from update_paths);
                insert or ignore into %(period)stally
                select path, type, 0 from update_paths;
                ''' % locals())
        self._conn.commit()
        
    def tallyPath(self, path, date, count=1):
        """Add count to path tally"""
        today = datetime.datetime.utcnow().date()
        daysago = (today - date).days
        periods = [p for p, l in self.periods.items() if daysago < l]
        if not periods:
            # path date is too long ago to be tallied
            return
        self._conn.execute('''\
            insert or replace into dailytally 
            values (:date, :path, :count + ifnull(
              (select count from dailytally 
               where path = :path and day = :date), 0))''', locals())
        for period in periods:
            self._conn.execute('update or ignore %stally '
                               'set count = count + :count '
                               'where path = :path' % period, locals())
        self._conn.commit()

def _test(module=None):
    import doctest
    return doctest.testmod(m=module)

if __name__ == '__main__':
    _test()
