try:
    set # C implementation in Python 2.4 and up
except NameError:
    from sets import Set as set
    
from itertools import ifilter, imap

def pathfilter(iterable, acceptable):
    """Filter (timestamp, path) iterable based on acceptible paths

      >>> items = ((1, '/foo'), (2, '/spam'), (3, '/bar'), (4, '/eggs'),
      ...          (5, '/spam'))
      >>> accepted = ('/spam', '/eggs')
      >>> tuple(pathfilter(items, accepted))
      ((2, '/spam'), (4, '/eggs'), (5, '/spam'))
    
    """
    acceptable = set(acceptable)
    def acceptablefilter(item):
        return item[1] in acceptable
    return ifilter(acceptablefilter, iterable)

def prefixfilter(iterable, prefix):
    """Filter (timestamp, path) iterable on paths starting with prefix
    
      >>> items = ((1, '/foo/bar'), (2, '/bar'))
      >>> tuple(prefixfilter(items, '/foo'))
      ((1, '/foo/bar'),)
    
    """
    def prefixfilter(item):
        return item[1].startswith(prefix)
    return ifilter(prefixfilter, iterable)
                
def removeprefix(iterable, prefix):
    """Remove prefix from paths in a (timestamp, path) iterable

      >>> items = ((1, '/foo/bar'), (2, '/foo/spam'), (3, '/bar'))
      >>> tuple(removeprefix(items, '/foo'))
      ((1, '/bar'), (2, '/spam'), (3, '/bar'))
    
    """
    def prefixremove(item):
        if item[1].startswith(prefix):
            return item[0], item[1][len(prefix):]
        return item
    return imap(prefixremove, iterable)

def removequery(iterable):
    """Remove the query parts of paths in a (timestamp, path) iterable
    
      >>> items = ((1, '/foo?spam=eggs&monty=python'), (2, '/bar'))
      >>> tuple(removequery(items))
      ((1, '/foo'), (2, '/bar'))
      
    """
    def queryremove(item):
        if '?' in item[1]:
            return item[0], item[1][:item[1].index('?')]
        return item
    return imap(queryremove, iterable)

def todate(iterable):
    """Convert timestamps in a (timestamp, path) iterable to dates

      >>> import datetime
      >>> day = datetime.timedelta(1)
      >>> now = datetime.datetime.now()
      >>> today = datetime.date.today()
      >>> items = ((now, '/foo'), (now - day, '/bar'))
      >>> result = tuple(todate(items))
      >>> result[0] == (today, '/foo')
      True
      >>> result[1] == (today - day, '/bar')
      True
    
    """
    def tstodate(item):
        return item[0].date(), item[1]
    return imap(tstodate, iterable)

class pathcollector(object):
    """Collect unique paths from a (path, type) iterable
    
      >>> items = (('/foo', 'type1'), ('/bar', 'type2'), ('/foo', 'type1'))
      >>> coll = pathcollector(items)
      >>> tuple(coll) == items
      True
      >>> paths = list(coll.paths)
      >>> paths.sort()
      >>> paths
      ['/bar', '/foo']
    
    """
    paths = None
    def __init__(self, iterable):
        self.paths = set()
        self._iterable = iterable
    def __iter__(self):
        for path, type in self._iterable:
            self.paths.add(path)
            yield path, type

try:
    from itertools import groupby # C implementation in Python 2.4 and up
except ImportError:
    class groupby(object):
        """groupby(iterable[, keyfunc]) -> create an iterator which returns
           (key, sub-iterator) grouped by each value of key(value).
        """
        def __init__(self, iterable, key=None):
            if key is None:
                key = lambda x: x
            self.keyfunc = key
            self.it = iter(iterable)
            self.tgtkey = self.currkey = self.currvalue = xrange(0)
        def __iter__(self):
            return self
        def next(self):
            while self.currkey == self.tgtkey:
                self.currvalue = self.it.next() # Exit on StopIteration
                self.currkey = self.keyfunc(self.currvalue)
            self.tgtkey = self.currkey
            return (self.currkey, self._grouper(self.tgtkey))
        def _grouper(self, tgtkey):
            while self.currkey == tgtkey:
                yield self.currvalue
                self.currvalue = self.it.next() # Exit on StopIteration
                self.currkey = self.keyfunc(self.currvalue)


def _test(module=None):
    import doctest
    return doctest.testmod(m=module)

if __name__ == '__main__':
    _test()
